import pathlib
import numpy as np
from   matplotlib.pyplot import imread as mpl_imread
from   .utils            import replace_all, dummy_function, name_from_file

class logviewer:
    """ log viewer
        Since lognflow makes lots of files and folders, maybe it is nice
        to have a logviewer that loads those information. In this module we
        provide a set of functions for a logged object that can load variables,
        texts, file lists and etc.. Use it simply by::
 
            from lognflow import logviewer
            logged = logviewer(log_dir = 'dir_contatining_files')
            var = logged.get_single('variable_name')
    """ 
    def __init__(self,
                 log_dir : pathlib.Path,
                 logger = print):
        self.log_dir = pathlib.Path(log_dir)
        assert self.log_dir.is_dir(), \
            f'lognflow.logviewer| No such directory: '+ str(self.log_dir)
        self.logger = logger
        self.load = self.get_single
        self.log_dir_str = str(self.log_dir.absolute())
    
    def name_from_file(self, fpath):
        """ 
            Given an fpath inside the logger log_dir, 
            what would be its equivalent parameter_name?
        """
        return name_from_file(self.log_dir_str, fpath)
    
    def disable_logger(self):
        self.logger = dummy_function
    
    def get_flist(self, var_name, suffix = None):
        """ get list of files
            return the list of files for a saved variable.

            Parameters
            ----------
            :param var_name:
                variable name
            :param suffix:
                If there are different suffixes availble for a variable
                this input needs to be set. npy, npz, mat, and torch are
                supported.
        """
        var_name = var_name.replace('\t', '\\t').replace('\n', '\\n')\
            .replace('\r', '\\r').replace('\b', '\\b')

        flist = list((self.log_dir).glob(var_name))
        
        if not flist:
            if suffix is None:
                if len(var_name.split('.')) > 1:
                    suffix = var_name.split('.')[-1]
                    name_before_suffix = var_name.split('.')[:-1]
                    if((len(name_before_suffix) == 1) & 
                       (name_before_suffix[0] == '')):
                        var_name = '*'
                    else:
                        var_name = ('.').join(var_name.split('.')[:-1])
                else:
                    suffix = '*'
    
            suffix = suffix.strip('.')        
    
            flist = []            
            if((self.log_dir / var_name).is_file()):
                flist = [self.log_dir / var_name]
            elif((self.log_dir / f'{var_name}.{suffix}').is_file()):
                flist = [self.log_dir / f'{var_name}.{suffix}']
            else:
                _var_name = (self.log_dir / var_name).name
                _var_dir = (self.log_dir / var_name).parent
                search_patt = f'{_var_name}*.{suffix}'
                search_patt = replace_all(search_patt, '**', '*')
                flist = list(_var_dir.glob(search_patt))
                if(len(flist) == 0):
                    search_patt = f'{_var_name}*.*'
                    search_patt = replace_all(search_patt, '**', '*')
                    flist = list(_var_dir.glob(search_patt))
                    if(len(flist) > 0):
                        self.logger(
                            'I Can not find the file with the given suffix, '\
                            + 'but found some with a different suffix, '\
                            + f'one file is: {flist[-1]}. This is what I'\
                            + ' will return.' )
                    
        if(flist):
            flist.sort()
        else:
            var_dir = self.log_dir / var_name
            if(var_dir.is_dir()):
                flist = list(var_dir.glob('*'))
            if(len(flist) > 0):
                flist.sort()
        return flist

    def get_common_files(self, var_name_A, var_name_B, suffix = None,
                               flist_A = None, flist_B = None):
        """ get common files in two directories
        
            It happens often in ML that there are two directories, A and B,
            and we are interested to get the flist in both that is common 
            between them. returns a tuple of two lists of files.
            
            Parameters
            ----------
            :param var_name_A:
                directory A name
            :param var_name_B:
                directory B name
        """
        if not flist_A:
            flist_A = self.get_flist(var_name_A, suffix)
        if not flist_B:
            flist_B = self.get_flist(var_name_B, suffix)
        
        suffix_A = flist_A[0].suffix
        suffix_B = flist_B[0].suffix 
        parent_A = flist_A[0].parent
        parent_B = flist_B[0].parent
        
        fstems_A = [_fst.stem for _fst in flist_A]
        fstems_B = [_fst.stem for _fst in flist_B]
        
        fstems_A_set = set(fstems_A)
        fstems_B_set = set(fstems_B)
        common_stems = list(fstems_A_set.intersection(fstems_B_set))

        flist_A_new = [parent_A / (common_stem + suffix_A) \
                          for common_stem in common_stems]
        flist_B_new = [parent_B / (common_stem + suffix_B) \
                          for common_stem in common_stems]

        return(flist_A_new, flist_B_new)

    # def load(self, parameter_name, suffix = None, flist = None,
    #              file_index : [int, list[int]] = -1, read_func = None):
    #     ...
    
    def get_text(self, log_name='main_log', flist = None, suffix = 'txt',
                       file_index = -1):
        """ get text log files
            Given the log_name, this function returns the text therein.

            Parameters
            ----------
            :param log_name:
                the log name. If not given then it is the main log.
            :param flist:
                you can give a file list in Posix paths, for text files
            :param suffix: str
                to search for specifi files
            :param file_index: int or list[int]
                a number or a list of numbers for the index of the file 
                to include, default: -1

        """
        if isinstance(file_index, int):
            file_index = [file_index]
        if not flist:
            flist = self.get_flist(log_name, suffix)
        n_files = len(flist)
        if (n_files>0):
            txt = []
            for fcnt in file_index:
                with open(flist[int(fcnt)]) as f_txt:
                    txt.append(f_txt.readlines())
            if(n_files == 1):
                txt = txt[0]
            return txt

    def _get_single(self, var_name, file_index = -1, 
                   suffix = None, read_func = None, verbose = False):
        """ get a single variable
            return the value of a saved variable.

            Parameters
            ----------
            :param var_name:
                variable name
            :param file_index:
                If there are many snapshots of a variable, this input can
                limit the returned to a set of indices.
            :param suffix:
                If there are different suffixes availble for a variable
                this input needs to be set. npy, npz, mat, and torch are
                supported.
            :param read_func:
                a function that takes the Posix path and returns data
            .. note::
                when reading a MATLAB file, the output is a dictionary.
                Also when reading a npz except if it is made by log_var
        """
        assert file_index == int(file_index), \
                    f'file_index {file_index} must be an integer'
        flist = self.get_flist(var_name, suffix)
        var_path = None
        if flist:
            if len(flist)>1:
                if verbose:
                    self.logger(  f'There are {len(flist)} files, logged with'
                                + f' name {var_name}.'
                                + f' The given index is {file_index}.')
            var_path = flist[file_index]
    
            if(var_path.is_file()):
                if verbose:
                    self.logger(f'Loading {var_path}')
                if read_func is not None:
                    return (read_func(var_path), var_path)
                if(var_path.suffix == '.npz'):
                    buf = np.load(var_path)
                    try: #check if it is made by log_var
                        assert len(buf.files) == 2
                        time_array = buf['time_array']
                        data_array = buf['data_array']
                        data_array = data_array[time_array > 0]
                        time_array = time_array[time_array > 0]
                        return((time_array, data_array), var_path)
                    except:
                        return(buf, var_path)
                if(var_path.suffix == '.npy'):
                    return(np.load(var_path), var_path)
                if(var_path.suffix == '.mat'):
                    from scipy.io import loadmat
                    return(loadmat(var_path), var_path)
                if(var_path.suffix == '.dm4'):
                    from hyperspy.api import load as hyperspy_api_load
                    return (hyperspy_api_load(var_path).data, var_path)
                if((var_path.suffix == '.tif') | (var_path.suffix == '.tiff')):
                    from tifffile import imread as tifffile_imread
                    return(tifffile_imread(var_path), var_path)
                if(var_path.suffix == '.torch'):      
                    from torch import load as torch_load 
                    return(torch_load(var_path), var_path)
                try:
                    img = mpl_imread(var_path)
                    return(img, var_path)
                except:
                    pass
                # if( (var_path.suffix in ['.txt', '.pdb', '.json', '.fasta'])):
                #     return(var_path.read_text(), var_path)
                try:
                    txt = var_path.read_text()
                    return(txt, var_path)
                except:
                    var_path = None
            else:
                var_path = None
                
        if (var_path is None) & verbose:
            self.logger(f'Looking for {var_name} failed. ' + \
                        f'{var_path} is not in: {self.log_dir}')
        return None, None
    
    def get_single(self, var_name, file_index = -1, 
                   suffix = None, read_func = None, verbose = False,
                   return_fpath = False):
        """ get a single variable
            return the value of a saved variable.

            Parameters
            ----------
            :param var_name:
                variable name
            :param file_index:
                If there are many snapshots of a variable, this input can
                limit the returned to a set of indices.
            :param suffix:
                If there are different suffixes availble for a variable
                this input needs to be set. npy, npz, mat, and torch are
                supported.
            :param read_func:
                a function that takes the Posix path and returns data
            .. note::
                when reading a MATLAB file, the output is a dictionary.
                Also when reading a npz except if it is made by log_var
        """
        get_single_data, fpath = self._get_single(
            var_name = var_name, file_index = file_index, suffix = suffix, 
            read_func = read_func, verbose = verbose)   
        if return_fpath:
            return get_single_data, fpath
        else:
            return get_single_data
        
    def get_stack_from_files(self, 
        var_name = None, flist = [], suffix = None, read_func = None):
       
        """ Get list or data of all files in a directory
       
            This function gives the list of paths of all files in a directory
            for a single variable.

            Parameters
            ----------
            :param var_name:
                The directory or variable name to look for the files
            :type var_name: str
           
            :param flist:
                list of Paths, if data is returned, this flist input can limit
                the data requested to this list.
            :type flist: list
           
            :param suffix:
                the suffix of files to look for, e.g. 'txt'
            :type siffix: str
           
            :param read_func:
                the function that takes the posix path of a file and returns
                the data in there.
           
            Output
            ----------
           
                It returns a list of data in all files or a numpy array if 
                concatenation of all is possible.
        """
        if not flist:
            flist = self.get_flist(var_name, suffix)
        else:
            flist = list(flist)
            assert pathlib.Path(flist[0]).is_file(), \
                f'File not found: {flist[0]}. You can use logviewer get_flist'
        
        if flist:
            n_files = len(flist)
            if(read_func is None):
                try:
                    fdata = np.load(flist[0])
                    read_func = np.load
                except:
                    pass
            if(read_func is None):
                try:
                    fdata = mpl_imread(flist[0])
                    read_func = mpl_imread
                except:
                    pass
            try:
                read_func(flist[0])
            except Exception as e:
                self.logger(f'The data file {flist[0]} could not be opened.'
                            'Please provide a read_function in the input.')
                raise e
            dataset = [read_func(fpath) for fpath in flist]
            try:
                dataset_array = np.array(dataset)
            except:
                dataset_array = dataset
            return(dataset_array)

    def get_stack_from_names(self, 
             var_names = None, read_func = None, return_flist = False):
        try:
            var_names_str = str(var_names)
        except:
            pass
        else:
            var_names = [var_names]
        assert var_names == list(var_names), \
            'input should be a list of variable names'
        dataset = []
        flist = []
        for name in var_names:
            images_flist = self.get_flist(name)
            if images_flist:
                for file_index in range(len(images_flist)):
                    data, fpath = self.get_single(
                        name, file_index = file_index,
                        read_func = read_func, return_fpath = True)
                    if data is not None:
                        dataset.append(data)
                        flist.append(fpath)
        if return_flist:
            return dataset, flist
        else:
            return dataset

    def replace_time_with_index(self, var_name):
        """ index in file var_names
            lognflow uses time stamps to make new log files for a variable.
            That is done by putting time stamp after the name of the variable.
            This function changes all of the time stamps, sorted ascendingly,
            by indices.
            
            :param var_name:
                variable name
        """
        var_dir = self.log_dir / var_name
        if(var_dir.is_dir()):
            var_fname = None
            flist = list(var_dir.glob(f'*.*'))
        else:
            var_fname = var_dir.name
            var_dir = var_dir.parent
            flist = list(var_dir.glob(f'{var_fname}_*.*'))
        if flist:
            flist.sort()
            fcnt_width = len(str(len(flist)))
            for fcnt, fpath in enumerate(flist):
                # self.log_text(None, f'Changing {flist[fcnt].name}')
                fname_new = ''
                if(var_fname is not None):
                    fname_new = var_fname + '_'
                fname_new += f'{fcnt:0{fcnt_width}d}' + flist[fcnt].suffix
                fpath_new = flist[fcnt].parent / fname_new
                # self.log_text(None, f'To {fpath_new.name}')
                flist[fcnt].rename(fpath_new)

    
    def __repr__(self):
        return f'{self.log_dir}'

    def __bool__(self):
        return self.log_dir.is_dir()


from .printprogress import printprogress
from mpl_toolkits.axes_grid1 import make_axes_locatable
import matplotlib.gridspec
import numpy as np
import matplotlib.pyplot as plt

def plt_colorbar(mappable):
    """ Add colobar to the current axis 
        This is specially useful in plt.subplots
        stackoverflow.com/questions/23876588/
            matplotlib-colorbar-in-each-subplot
    """
    ax = mappable.axes
    fig = ax.figure
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    cbar = fig.colorbar(mappable, cax=cax)
    return cbar

def plt_hist(vectors_list, 
             n_bins = 10, alpha = 0.5, normalize = False, 
             labels_list = None, **kwargs):
    
    fig = plt.figure()
    ax = fig.add_subplot(111)
    if not (type(vectors_list) is list):
        vectors_list = [vectors_list]
    for vec_cnt, vec in enumerate(vectors_list):
        bins, edges = np.histogram(vec, n_bins)
        if normalize:
            bins = bins / bins.max()
        ax.bar(edges[:-1], bins, 
                width =np.diff(edges).mean(), alpha=alpha)
        if labels_list is None:
            ax.plot(edges[:-1], bins, **kwargs)
        else:
            assert len(labels_list) == len(vectors_list)
            ax.plot(edges[:-1], bins, 
                     label = f'{labels_list[vec_cnt]}', **kwargs)
    return fig, ax

def plt_surface(stack, **kwargs):
    from mpl_toolkits.mplot3d import Axes3D
    n_r, n_c = stack.shape
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    X, Y = np.meshgrid(np.arange(n_r, dtype='int'), 
                       np.arange(n_c, dtype='int'))
    ax.plot_surface(X, Y, stack, **kwargs)
    return fig, ax
        
def pltfig_to_numpy(fig):
    """ from https://www.icare.univ-lille.fr/how-to-
                    convert-a-matplotlib-figure-to-a-numpy-array-or-a-pil-image/
    """
    fig.canvas.draw()
    w,h = fig.canvas.get_width_height()
    buf = np.frombuffer(fig.canvas.tostring_argb(), dtype=np.ubyte)
    buf.shape = (w, h, 4)
    return buf.sum(2)

def numbers_as_images_3D(data3D_shape: tuple,
                         fontsize: int, 
                         text_loc: tuple = None,
                         verbose: bool = True):
    """ Numbers3D
    This function generates a 4D dataset of images with shape
    (n_x, n_r, n_c) where in each image the value "x" is written as a text
    that fills the image. As such, later when working with such a dataset you can
    look at the image and know which index it had before you use it.
    
    Follow this recipe to make good images:
    
    1- set n_x to 10, Set the desired n_r, n_c and width. 
    2- find fontsize that is the largest and still fits
    3- Increase n_x to desired size.
    
    You can provide a logs_root, log_dir or simply select a directory to save the
    output 3D array.
    
    """
    n_x, n_r, n_c = data3D_shape
    
    if text_loc is None:
        text_loc = (n_r//2 - fontsize, n_c//2 - fontsize)
    
    dataset = np.zeros(data3D_shape)    
    txt_width = int(np.log(n_x)/np.log(n_x)) + 1
    number_text_base = '{ind_x:0{width}}}'
    if(verbose):
        pBar = printprogress(n_x)
    for ind_x in range(n_x):
        mat = np.ones((n_r, n_c))
        number_text = number_text_base.format(ind_x = ind_x, 
                                              width = txt_width)
        fig = plt.figure(figsize = (n_rr, n_cc), dpi = n_rc)
        ax = fig.add_subplot(111)
        ax.imshow(mat, cmap = 'gray', vmin = 0, vmax = 1)
        ax.text(text_loc[0], text_loc[1],
                number_text, fontsize = fontsize)
        ax.axis('off')
        buf = pltfig_to_numpy(fig)
        plt.close()
        dataset[ind_x] = buf.copy()
        if(verbose):
            pBar()
    return dataset

def numbers_as_images_4D(data4D_shape: tuple,
                         fontsize: int, 
                         text_loc: tuple = None,
                         verbose: bool = True):
    """ Numbers4D
    This function generates a 4D dataset of images with shape
    (n_x, n_y, n_r, n_c) where in each image the value "x, y" is written as a text
    that fills the image. As such, later when working with such a dataset you can
    look at the image and know which index it had before you use it.
    
    Follow this recipe to make good images:
    
    1- set n_x, n_y to 10, Set the desired n_r, n_c and width. 
    2- try fontsize that is the largest
    3- Increase n_x and n_y to desired size.
    
    You can provide a logs_root, log_dir or simply select a directory to save the
    output 4D array.
    
    :param text__loc:
        text_loc should be a tuple of the location of bottom left corner of the
        text in the image.
    
    """
    n_x, n_y, n_r, n_c = data4D_shape

    if text_loc is None:
        text_loc = (n_r//2 - fontsize, n_c//2 - fontsize)
    
    dataset = np.zeros((n_x, n_y, n_r, n_c))    
    txt_width = int(np.log(np.maximum(n_x, n_y))
                    / np.log(np.maximum(n_x, n_y))) + 1
    number_text_base = '{ind_x:0{width}}, {ind_y:0{width}}'
    if(verbose):
        pBar = printprogress(n_x * n_y)
    for ind_x in range(n_x):
        for ind_y in range(n_y):
            mat = np.ones((n_r, n_c))
            number_text = number_text_base.format(
                ind_x = ind_x, ind_y = ind_y, width = txt_width)
            n_rc = np.minimum(n_r, n_c)
            n_rr = n_r / n_rc
            n_cc = n_c / n_rc
            fig = plt.figure(figsize = (n_rr, n_cc), dpi = n_rc)
            ax = fig.add_subplot(111)
            ax.imshow(mat, cmap = 'gray', vmin = 0, vmax = 1)
            ax.text(text_loc[0], text_loc[1], number_text, fontsize = fontsize)
            ax.axis('off')
            buf = pltfig_to_numpy(fig)
            plt.close()
            dataset[ind_x, ind_y] = buf.copy()
            if(verbose):
                pBar()
    return dataset

class plot_gaussian_gradient:
    """ Orignally developed for RobustGaussinFittingLibrary
    Plot curves by showing their average, and standard deviatoin
    by shading the area around the average according to a Gaussian that
    reduces the alpha as it gets away from the average.
    You need to init() the object then add() plots and then show() it.
    refer to the tests.py
    """
    def __init__(self, xlabel = None, ylabel = None, num_bars = 100, 
                 title = None, xmin = None, xmax = None, 
                 ymin = None, ymax = None, fontsize = 14):
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.title = title
        self.num_bars = num_bars
        self.xmin = xmin
        self.xmax = xmax
        self.ymin = ymin
        self.ymax = ymax
        LWidth = 1
        font = {
                'weight' : 'bold',
                'size'   : fontsize}
        plt.rc('font', **font)
        params = {'legend.fontsize': 'x-large',
                 'axes.labelsize': 'x-large',
                 'axes.titlesize':'x-large',
                 'xtick.labelsize':'x-large',
                 'ytick.labelsize':'x-large'}
        plt.rcParams.update(params)
        plt.figure(figsize=(8, 6), dpi=50)
        self.ax1 = plt.subplot(111)
    
    def addPlot(self, x, mu, std, gradient_color, label, 
                snr = 3.0, mu_color = None, general_alpha = 1,
                mu_linewidth = 1):

        for idx in range(self.num_bars-1):
            y1 = ((self.num_bars-idx)*mu + idx*(mu + snr*std))/self.num_bars
            y2 = y1 + snr*std/self.num_bars
            
            prob = np.exp(-(snr*idx/self.num_bars)**2/2)
            plt.fill_between(
                x, y1, y2, 
                color = (gradient_color + (prob*general_alpha,)), 
                edgecolor=(gradient_color + (0,)))

            y1 = ((self.num_bars-idx)*mu + idx*(mu - snr*std))/self.num_bars
            y2 = y1 - snr*std/self.num_bars
            
            plt.fill_between(
                x, y1, y2, 
                color = (gradient_color + (prob*general_alpha,)), 
                edgecolor=(gradient_color + (0,)))
        if(mu_color is None):
            mu_color = gradient_color
        plt.plot(x, mu, linewidth = mu_linewidth, color = mu_color, 
                 label = label)
        
    def show(self, show_legend = True):
        if(self.xmin is not None) & (self.xmax is not None):
            plt.xlim([self.xmin, self.xmax])
        if(self.ymin is not None) & (self.ymax is not None):
            plt.ylim([self.ymin, self.ymax])
        if(self.xlabel is not None):
            plt.xlabel(self.xlabel, weight='bold')
        if(self.ylabel is not None):
            plt.ylabel(self.ylabel, weight='bold')
        if(self.title is not None):
            plt.title(self.title)
        if(show_legend):
            plt.legend()
        plt.grid()
        
        plt.show()

def imshow_series(list_of_stacks, 
                  list_of_masks = None,
                  figsize_ratio = 1,
                  text_as_colorbar = False,
                  colorbar = False,
                  cmap = 'viridis',
                  list_of_titles = None,
                  ):
    """ imshow a stack of images or sets of images in a shelf
        input must be a list or array of images
        
        Each element of the list can appear as either:
        * n_im, n_r x n_c
        * n_im, n_r x  3  x 1
        * n_im, n_r x n_c x 3

        :param list_of_stacks
                list_of_stacks would include arrays iteratable by their
                first dimension.
        :param borders: float
                borders between tiles will be filled with this variable
                default: np.nan
    """
    n_stacks = len(list_of_stacks)
    if(list_of_masks is not None):
        assert len(list_of_masks) == n_stacks, \
            f'the number of masks, {len(list_of_masks)} and ' \
            + f'stacks {n_stacks} should be the same'
    
    # if list_of_titles is not None:
    #     assert len(list_of_titles) == n_stacks, \
    #         f'the number of titles, {len(list_of_titles)} and ' \
    #         + f'stacks {n_stacks} should be the same'
    
    n_imgs = list_of_stacks[0].shape[0]
    for ind, stack in enumerate(list_of_stacks):
        assert stack.shape[0] == n_imgs, \
            'All members of the given list should have same number of images.'
        assert (len(stack.shape) == 3) | (len(stack.shape) == 4), \
            f'The shape of the stack {ind} must have length 3 or 4, it has '\
            f' shape of {stack.shape}. Perhaps you wanted to have only one set of'\
            ' images. If thats the case, put that single image in a list.'
            
    fig = plt.figure(figsize = (n_imgs*figsize_ratio,n_stacks*figsize_ratio))
    gs1 = matplotlib.gridspec.GridSpec(n_stacks, n_imgs)
    if(colorbar):
        gs1.update(wspace=0.25, hspace=0)
    else:
        gs1.update(wspace=0.025, hspace=0) 
    
    for img_cnt in range(n_imgs):
        for stack_cnt in range(n_stacks):
            ax = plt.subplot(gs1[stack_cnt, img_cnt])
            plt.axis('on')
            ax.set_xticklabels([])
            ax.set_yticklabels([])
            data_canvas = list_of_stacks[stack_cnt][img_cnt].copy()
            if(list_of_masks is not None):
                mask = list_of_masks[stack_cnt]
                if(mask is not None):
                    if(data_canvas.shape == mask.shape):
                        data_canvas[mask==0] = 0
                        data_canvas_stat = data_canvas[mask>0]
            else:
                data_canvas_stat = data_canvas.copy()
            data_canvas_stat = data_canvas_stat[
                np.isnan(data_canvas_stat) == 0]
            data_canvas_stat = data_canvas_stat[
                np.isinf(data_canvas_stat) == 0]
            vmin = data_canvas_stat.min()
            vmax = data_canvas_stat.max()
            im = ax.imshow(data_canvas, 
                            vmin = vmin, 
                            vmax = vmax,
                            cmap = cmap)
            if(text_as_colorbar):
                ax.text(data_canvas.shape[0]*0,
                         data_canvas.shape[1]*0.05,
                         f'{data_canvas.max():.6f}', 
                         color = 'yellow',
                         fontsize = 2)
                ax.text(data_canvas.shape[0]*0,
                         data_canvas.shape[1]*0.5, 
                         f'{data_canvas.mean():.6f}', 
                         color = 'yellow',
                         fontsize = 2)
                ax.text(data_canvas.shape[0]*0,
                         data_canvas.shape[1]*0.95, 
                         f'{data_canvas.min():.6f}', 
                         color = 'yellow',
                         fontsize = 2)
            if(colorbar):
                cbar = plt.colorbar(im, ax=ax, fraction=0.046, pad=0.04)
                cbar.ax.tick_params(labelsize=1)
            ax.set_aspect('equal')
            ax.axis('off')
        
    
    
    return fig, None

def imshow_by_subplots( 
    stack,
    frame_shape = None,
    grid_locations = None,
    figsize = None,
    im_size_factor = None,
    im_sizes = None,
    colorbar = False,
    remove_axis_ticks = True,
    title = None,
    cmap = None,
    **kwargs):    
    
    stack_shape = stack.shape
    n_dims = len(stack_shape)
    
    FLAG_img_ready = False
    use_stack_to_frame = False
    if(n_dims == 2):
        FLAG_img_ready = True
    elif(n_dims == 3):
        if(stack_shape[2] != 3):
            use_stack_to_frame = True
        else:
            #warning that 3 dimensions as the last axis is RGB
            FLAG_img_ready = True
    elif(n_dims == 4):
            use_stack_to_frame = True
    
    if(use_stack_to_frame):
        FLAG_img_ready = True

    if(FLAG_img_ready):
        if(np.iscomplexobj(stack)):
            print('complex not supported in log_imshow_by_subplots')
            return
        else:                
            n_f = stack.shape[0]
            if grid_locations is None:
                if(frame_shape is None):
                    n_f_sq = int(np.ceil(n_f ** 0.5))
                    n_f_r, n_f_c = (n_f_sq, n_f_sq)
                else:
                    n_f_r, n_f_c = frame_shape
                
                fig, ax = plt.subplots(n_f_r,n_f_c)
                if(remove_axis_ticks):
                    plt.setp(ax, xticks=[], yticks=[])
                for rcnt in range(n_f_r):
                    for ccnt in range(n_f_c):
                        imcnt = ccnt + rcnt * n_f_c
                        if imcnt < n_f:
                            im = stack[imcnt]
                            im_ch = ax[rcnt, ccnt].imshow(
                                im, cmap = cmap, **kwargs)
                            if(colorbar):
                                plt_colorbar(im)
                            if(remove_axis_ticks):
                                plt.setp(ax, xticks=[], yticks=[])
            else:
                assert len(grid_locations) == n_f, \
                    f'length of grid_locations: {grid_locations.shape} should '\
                    f'be the same as number of images: {n_f}.'
                assert len(grid_locations.shape) == 2, \
                    'grid_locations should be n_f x 2, its shape is: '\
                    f'{grid_locations.shape}.'
                if background_image is not None:
                    background_image = background_image.squeeze()
                    assert len(background_image.shape) == 2, \
                        'The background image should be a 2D image, its shape ' \
                        f' is {background_image.shape}.'
                        
                if figsize is None:
                    grid_locations_r_min = grid_locations[:, 0].min()
                    grid_locations_r_max = grid_locations[:, 0].max()
                    grid_locations_c_min = grid_locations[:, 1].min()
                    grid_locations_c_max = grid_locations[:, 1].max()
                    grid_size = (grid_locations_r_max - grid_locations_r_min,
                                 grid_locations_c_max - grid_locations_c_min)
                    figsize = (2, 2 * grid_size[1]/grid_size[0])
                if im_sizes is None:
                    if im_size_factor is None:
                        im_size_factor = figsize[0] * figsize[1] / n_f
                        im_sizes = (im_size_factor, 
                                    im_size_factor * stack.shape[
                                        2]/stack.shape[1]) 
                
                fig = plt.figure(figsize=figsize)
                ax = fig.add_axes([0, 0, 1, 1])
                if background_image is not None:
                    ax.imshow(background_image)
                if title is not None:
                    ax.set_title(title)
                for ccnt, coords in enumerate(grid_locations):
                    pos = [coords[1], 1-coords[0], im_sizes[0], im_sizes[1]]
                    ax_local = fig.add_axes(pos)
                    im = ax_local.imshow(stack[ccnt], 
                                   cmap = cmap, **kwargs)
                    if(colorbar):
                        plt_colorbar(im)
                    if(remove_axis_ticks):
                        plt.setp(ax_local, xticks=[], yticks=[])
                        ax_local.xaxis.set_ticks_position('none')
                        ax_local.yaxis.set_ticks_position('none')
        
        return fig, ax
    else:
        self.log_text(
            self.log_name,
            f'Cannot imshow variable {parameter_name} with shape' + \
            f'{stack.shape}')
        return


#!/usr/bin/env python

"""Tests for `lognflow.logviewer` package."""

import pytest
import re
import numpy as np

from lognflow import (lognflow, select_directory, 
                      logviewer, printprogress,
                      text_to_object)

import tempfile
temp_dir = tempfile.gettempdir()

def test_get_flist_multiple_directories():
    logger = lognflow(temp_dir)
    logger('Well this is a test for test_multiple_directories_get_flist')
    
    logger.log_single('dir1/dir/var', np.random.rand(100))
    logger.log_single('dir2/dir/var', np.random.rand(100))
    logger.log_single('dir3/dir/var', np.random.rand(100))
    
    flist = logger.logged.get_flist('dir*/dir/var*.npy')
    [print(_) for _ in flist]
    [print(logger.logged.name_from_file(_)) for _ in flist]
        
def test_logviewer():
    logger = lognflow(temp_dir)
    logger('Well this is a test for logviewer')
    
    logger.log_single('test_param', np.random.rand(100))
    
    logged = logviewer(logger.log_dir, logger)
    print(logged.get_single('test_param'))
    print(logged.get_text())

def test_get_stack_from_files():
    logger = lognflow(temp_dir)
    
    logger('Well this is a test for logviewer')

    for _ in range(5):
        logger.log_single('A/', np.random.rand(100, 100))
        logger.log_single('B/', np.random.randn(100, 100))

    logged = logviewer(logger.log_dir, logger)

    flist_A = logged.get_flist('A/*')
    flist_B = logged.get_flist('B/*')
    
    logger(flist_A)
    logger(flist_B)
    
    logger.logged.replace_time_with_index('A/*')
    logger.logged.replace_time_with_index('B/*')

    flist_A = logged.get_flist('A/*')
    flist_B = logged.get_flist('B/*')
    
    logger(flist_A)
    logger(flist_B)
    
    stack_A = logged.get_stack_from_files(flist = flist_A)
    stack_B = logged.get_stack_from_files(flist = flist_B)

    logger(f'stack_A.shape: {stack_A.shape}')
    logger(f'stack_B.shape: {stack_B.shape}')
    
    logger.log_imshow_series('data_samples', [stack_A, stack_B], dpi = 300)

    flist_A_AB, flist_B_AB = logged.get_common_files('A/*', 'B/*')
    logger(flist_A_AB)
    logger(flist_B_AB)
    
    if(flist_A_AB):
        
        dataset_A = logged.get_stack_from_files('A/*', flist = flist_A_AB)
        dataset_B = logged.get_stack_from_files('B/*', flist = flist_B_AB)
        
        logger.log_imshow_series('data_samples', 
                                 [dataset_A, dataset_B], dpi = 300)
        _ = logger._loggers_dict['main_log.txt'].log_size
        logger('Size of the log file in bytes is: ' \
               + f'{_}')

def test_text_to_object():
    logger = lognflow(temp_dir, time_tag = False)
    test_list = ['asdf', 1243, "dd"]
    logger.log_single('test_list', test_list, suffix = 'txt')
    
    test_dict = {"one": "asdf", 'two': 1243, 'thre': "dd"}
    logger.log_single('test_dict', test_dict, suffix = 'txt')
    
    logged = logviewer(logger.log_dir)
    flist = logged.get_flist('*')
    print(flist)
    for file_name_input in flist:
        print('='*60)
        print(f'file name: {file_name_input}')
        with open(file_name_input, 'r') as opened_txt:
            txt = opened_txt.read()
        print('text read from the file:')
        print(txt)
        print('- '*30)
        ext_obj = text_to_object(txt)
        print(f'Extracted object is of type {type(ext_obj)}:')
        print(ext_obj)

def test_get_single_specific_fname():
    logger = lognflow(temp_dir)
    logger('test get single specific fname')
    
    vec = np.array([1])
    logger.log_single('vec', vec, time_tag = False)

    vec2 = np.array([2])
    logger.log_single('vec2', vec2, time_tag = False)
    
    logged = logviewer(logger.log_dir)
    vec_out = logged.get_single('vec.npy')
    
    assert vec_out == vec

def test_get_stack_from_names():
    logger = lognflow(temp_dir)
    logger('test get single specific fname')
    
    logger.log_imshow('im1', np.random.randn(30, 30))
    logger.log_imshow('im1', np.random.randn(20, 40))
    
    logged = logviewer(logger.log_dir)

    images = logged.get_stack_from_names('im1')
    
    print(len(images))

if __name__ == '__main__':
    temp_dir = select_directory()
    test_get_stack_from_files()
    test_get_stack_from_names()
    test_get_flist_multiple_directories()
    test_get_single_specific_fname()
    test_logviewer()
    test_text_to_object()



from math import ceil
from time import time

class printprogress:
    """
    While there are packages that use \r to show a progress bar, 
    there are cases e.g. a print_function or an ssh terminal that does not 
    support \r. In such cases, if something is typed at the end of 
    the line, it cannot be deleted. The following code provides a good 
    looking progress bar with a simple title and limits and 
    is very simple to use. Define the object with number of steps of the loop 
    and then call it in every iteration of the loop. If you'd like it
    to go faster, call it with give number of steps that have passed.
    """
    def __init__(self, 
                 it, 
                 numTicks = 78,
                 title = None,
                 method = 'linear',
                 print_function = print,
                 **print_function_kwargs):
        """
            it: int or iterable
                Number of iterations in the for loop or iterable
                n_steps = int(it) or n_steps = len(it)
            numTicks: int
                The number of charachters in a row of the screen - 2
                default: 78 for old screens that have 80 coloumns
            title : str 
                The title of progress bar.
                default: f'Progress bar for {n_steps} steps in {numTicks} ticks'
            print_function:
                print_function must be callable with a string and should not add
                \n to the its input.
                If you pass None as the print_function, nothing will be printed.
                yet the output of the __call__ will be the remaining time in
                seconds.
            method: options how to calculate the remaining time are
                'linear':
                       p                   x      
                |-----------|--------------------------|
                           ck                       n_steps
                As such  x/(n-c) = p/c => x = p(n/c - 1)
                more options to come
        """
        assert method in ['linear', 'linear_robust']
        
        try:
            n_steps = int(it)
            self.yielding_data = False
        except:
            n_steps = len(it)
            self.it = it
            self.yielding_data = True
            self.yielding_data_call_warning = False
        self.FLAG_first_tick = True
        
        self.print_function_kwargs = print_function_kwargs
        self.method = method
        self.in_print_function = print_function
        
        if(n_steps<2):
            n_steps = 2
        if (title is None):
            title = f'Progress for {n_steps} steps'
        self.FLAG_ended = False
        self.FLAG_warning = False
        self.startTime = time()
        self.ck = 0
        self.prog = 0
        self.n_steps = n_steps
        if(numTicks < len(title) + 2 ):
            self.numTicks = len(title)+2
        else:
            self.numTicks = numTicks
        
        self._print_func(' ', end='')
        self._print_func('_'*self.numTicks, end='')
        self._print_func(' ')
        
        self._print_func('/', end='')
        self._print_func(' '*int((self.numTicks - len(title))/2), end='')
        self._print_func(title, end='')
        self._print_func(' '*int(ceil((self.numTicks-len(title))/2)-1), end='')
        self._print_func(' \\')
        
        self._print_func(' ', end = '')
        self.len_prog_text = 0
        self.remTimeS_perv = 0
        self.average_filter_coeff = 0
    
    def _print_func(self, text, end='\n'):
        if (self.in_print_function is not None):
            if (self.in_print_function == print):
                print(text, end = end, flush = True)
            else:
                self.in_print_function(text, end = end,
                                       **self.print_function_kwargs)
        
    def _calc_ETA(self):
        if(self.method == 'linear'):
            passedTime = time() - self.startTime
            if self.ck > 0:
                remTimeS = passedTime * ( self.n_steps / self.ck - 1)
                if(self.average_filter_coeff):
                    if (self.remTimeS_perv > 5) & (remTimeS > 5):
                        remTimeS = (1 - self.average_filter_coeff)*remTimeS + \
                                    self.average_filter_coeff*self.remTimeS_perv
                
                self.remTimeS_perv = remTimeS
            else:
                remTimeS = 1e+7
        return remTimeS
    
    def _make_progress(self, ck = 1):
        
        remTimeS = 0
        if(self.FLAG_ended):
            if(not self.FLAG_warning):
                self.FLAG_warning = True
                self._print_func('-' * (self.numTicks + 2))
        else:
            self.ck += ck
            if(self.ck <= self.n_steps):
                remTimeS = self._calc_ETA() # useful when print_function is None
                cProg = int(self.numTicks*self.ck/(self.n_steps-1)/3)
                #3: because 3 charachters are used
                while((self.prog < cProg) & (not self.FLAG_ended)):
                    self.prog += 1
                    remTimeS = self._calc_ETA()
                    if remTimeS < 86400*100:
                        if(remTimeS>356400): # less than 99d and more than 99h
                            progStr = "%02d" % int(ceil(remTimeS/86400))
                            self._print_func(progStr, end='')
                            self._print_func('d', end='')
                            self.len_prog_text += 3
                        elif(remTimeS>5940): # less than 99h and more than 99m
                            progStr = "%02d" % int(ceil(remTimeS/3600))
                            self._print_func(progStr, end='')
                            self._print_func('h', end='')
                            self.len_prog_text += 3
                        elif(remTimeS>99): # less than 99m and more than 99s
                            progStr = "%02d" % int(ceil(remTimeS/60))
                            self._print_func(progStr, end='')
                            self._print_func('m', end='')
                            self.len_prog_text += 3
                        elif(remTimeS>=0): # less than 99s and more than 0
                            progStr = "%02d" % int(ceil(remTimeS))
                            self._print_func(progStr, end='')
                            self._print_func('s', end='')
                            self.len_prog_text += 3
                        else:
                            self._end()
            if((self.ck >= self.n_steps) | 
               (self.len_prog_text >= self.numTicks)):
                self._end()
        return remTimeS
    
    def __call__(self, ck=1):
        """ ticking the progress bar
            just call the object and the progress bar moves ck steps
            ahead when ready.
            
            output
            ~~~~~~
            :param ETA:
                the remaining time in seconds will be provided at the output
        """
        if(self.yielding_data):
            if not self.yielding_data_call_warning:
                print('printprogress is used as an Iterator,'
                      ' You can not call it.')
                self.yielding_data_call_warning = True
        else:
            return self._make_progress(ck)

    # Supporting iterator type usage
    def __iter__(self):
        self.FLAG_iter_ended = False
        self.iter_ck = 0
        return self
  
    def __next__(self):
        if(self.yielding_data):
            if not self.FLAG_iter_ended:
                if self.FLAG_first_tick:
                    self._make_progress(0)
                    self.FLAG_first_tick = False
                else:
                    self._make_progress()
                if self.iter_ck == self.n_steps - 1:
                    self.FLAG_iter_ended = True
                toret = self.it[self.iter_ck]
                self.iter_ck += 1
                return toret
            else:
                raise StopIteration
    
    def _end(self):
        if(not self.FLAG_ended):
            self._print_func('')
            self.FLAG_ended = True

    def __del__(self):
        self._end()


"""Top-level package for lognflow."""

__author__ = 'Alireza Sadri'
__email__ = 'arsadri@gmail.com'
__version__ = '0.10.5'

from .lognflow import lognflow
from .logviewer import logviewer
from .printprogress import printprogress
from .plt_utils import plt_colorbar, plot_gaussian_gradient
from .utils import (
    select_directory, select_file, repr_raw, replace_all, 
    text_to_object, stack_to_frame, stacks_to_frames, ssh_system)
from .multiprocessor import multiprocessor
from .loopprocessor import loopprocessor


from lognflow import multiprocessor, loopprocessor, printprogress
import numpy as np
import inspect

def multiprocessor_targetFunc(iterables_sliced, shareables):
    idx = iterables_sliced[0]
    data, mask, op_type = shareables
    _data = data[idx]
    if(op_type=='median'):
        to_return1 = np.median(_data[mask[idx]==1])
        to_return1 = np.array([to_return1])
    to_return2 = np.ones((int(10*np.random.rand(1)), 2, 2))
    
    return(to_return1, 'median', to_return2)
    
def test_multiprocessor():
    print('-'*80, '\n', inspect.stack()[0][3], '\n', '-'*80)

    N = 10000
    D = 1000
    data = (10+100*np.random.randn(N,D)).astype('int')
    mask = (2*np.random.rand(N,D)).astype('int')
    op_type = 'median'

    shareables  = (data, mask, op_type)
    iterables = N
    
    stats = multiprocessor(multiprocessor_targetFunc, 
                           iterables, shareables,
                           verbose = True)

    results = []
    for cnt in range(N):
        results.append(multiprocessor_targetFunc((cnt, ), shareables))

    medians, otherOutput, _ids = stats
    print('type(medians)', type(medians))
    print('medians.shape', medians.shape)
    print('type(otherOutput)', type(otherOutput))
    print('len(otherOutput)', len(otherOutput))
    print('otherOutput[1] ', otherOutput[1])
    print('otherOutput[1][0] ', otherOutput[1][0])
    print('type(_ids) ', type(_ids))
    print('len(_ids) ', len(_ids))
    print('type(_ids[0]) ', type(_ids[0]))
    print('_ids.shape ', _ids.shape)
    
    direct_medians = np.zeros(N)
    for cnt in range(N):
        direct_medians[cnt] = np.median(data[cnt, mask[cnt]==1])
    
    print(np.array([ medians, direct_medians] ).T)
    print('difference of results: ', (direct_medians - medians).sum())

def masked_cross_correlation(iterables_sliced, shareables):
    vec1, vec2 = iterables_sliced
    mask, statistics_func = shareables
    vec1 = vec1[mask==1]
    vec2 = vec2[mask==1]
    
    vec1 -= vec1.mean()
    vec1_std = vec1.std()
    if vec1_std > 0:
        vec1 /= vec1_std
    vec2 -= vec2.mean()
    vec2_std = vec2.std()
    if vec2_std > 0:
        vec2 /= vec2_std

    correlation = vec1 * vec2
    to_return = statistics_func(correlation)
    return(to_return)

def test_multiprocessor_ccorr():
    print('-'*80, '\n', inspect.stack()[0][3], '\n', '-'*80)
    data_shape = (1000, 2000)
    data1 = np.random.randn(*data_shape)
    data2 = 2 + 5 * np.random.randn(*data_shape)
    mask = (2*np.random.rand(data_shape[1])).astype('int')
    statistics_func = np.median
    
    iterables = (data1, data2)
    shareables = (mask, statistics_func)
    ccorr = multiprocessor(
        masked_cross_correlation, iterables, shareables,
        test_mode = False)

def error_multiprocessor_targetFunc(iterables_sliced, shareables):
    idx = iterables_sliced[0]
    data, mask, op_type = shareables
    _data = data[idx]
    if(op_type=='median'):
        to_return1 = np.median(_data[mask[idx]==1])
        to_return1 = np.array([to_return1])
    to_return2 = np.ones((int(10*np.random.rand(1)), 2, 2))
    
    if idx == 3000:
        raise ValueError
    
    return(to_return1, 'median', to_return2)    

def test_error_handling_in_multiprocessor():
    print('-'*80, '\n', inspect.stack()[0][3], '\n', '-'*80)
    
    N = 10000
    D = 1000
    data = (10+100*np.random.randn(N,D)).astype('int')
    mask = (2*np.random.rand(N,D)).astype('int')
    op_type = 'median'

    iterables = N
    shareables  = (data, mask, op_type)
    
    stats = multiprocessor(
        error_multiprocessor_targetFunc, iterables, shareables,
        verbose = True)
    
def noslice_multiprocessor_targetFunc(iterables_sliced, shareables):
    idx = iterables_sliced
    data, mask, op_type = shareables
    _data = data[idx]
    if(op_type=='median'):
        to_return1 = np.median(_data[mask[idx]==1])
        to_return1 = np.array([to_return1])
    to_return2 = np.ones((int(10*np.random.rand(1)), 2, 2))
    return(to_return1, 'median', to_return2)    

def test_noslice_multiprocessor():
    print('-'*80, '\n', inspect.stack()[0][3], '\n', '-'*80)
    
    N = 10000
    D = 1000
    data = (10+100*np.random.randn(N,D)).astype('int')
    mask = (2*np.random.rand(N,D)).astype('int')
    op_type = 'median'

    iterables = N
    shareables  = (data, mask, op_type)
    
    stats = multiprocessor(
        noslice_multiprocessor_targetFunc, iterables, shareables, verbose = True)

def compute(data, mask):
    for _ in range(400):
        res = np.median(data[mask==1])
        
    # if (data>130).sum() > 0:
    #     asdf
        
    return res

def test_loopprocessor():
    print('-'*80, '\n', inspect.stack()[0][3], '\n', '-'*80)

    N = 16
    D = 1000000
    data = (100+10*np.random.randn(N,D)).astype('int')
    mask = (2*np.random.rand(N,D)).astype('int')

    print('-'*80)
    print('beggining of loop processor')
    print('-'*80)
    compute_mp = loopprocessor(compute)
    for cnt in printprogress(range(N)):
        compute_mp(data[cnt], mask[cnt])
    results_mp = compute_mp()

    results = np.zeros(N)
    for cnt in printprogress(range(N)):
        results[cnt] = compute(data[cnt], mask[cnt])

    print((results - results_mp).sum())
    
if __name__ == '__main__':
    test_loopprocessor()
    test_noslice_multiprocessor()
    test_multiprocessor()
    test_multiprocessor_ccorr()
    print('lets test', flush=True)
    try:
        test_error_handling_in_multiprocessor()
    except Exception as e:
        print('Error handled!')
        print('Error was:')
        print(e)



#!/usr/bin/env python

"""The setup script for lognflow."""

from setuptools import setup, find_packages

__author__ = 'Alireza Sadri'
__email__ = 'arsadri@gmail.com'
__version__ = '0.10.5'

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = ['numpy', 'matplotlib']

test_requirements = ['pytest>=3', ]

setup(
    author=__author__,
    author_email=__email__,
    python_requires='>=3.7',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11'
    ],
    description="Log and Flow tracking made easy with Python",
    install_requires=requirements,
    license="GNU General Public License v3",
    long_description = readme + '\n\n' + history,
    long_description_content_type = 'text/markdown',
    include_package_data=True,
    keywords='lognflow',
    name='lognflow',
    packages=find_packages(include=['lognflow', 'lognflow.*']),
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/arsadri/lognflow',
    version=__version__,
    zip_safe=False,
)